# BitKeeper syntax changes for tcl
#
# Copyright (c) 2005, Larry McVoy
# %K% 
#
# this file defines the following bits of goodness:
# struct <name> <definition>
#   - create a c-style struct 
# new <struct-name> ?instance-name?
#   - creates new instance of the given struct
# scope name body
#   - executes code in the named scope. All procs defined within
#     the scope automagically inherit scope variables; non-private
#     procs are made available to the global namespace. Code within
#     a scope can also use the pointer-to goodness
# scope_code arg ?arg arg ...?
#   - return a fully qualified command that can be run outside the
#     current scope
# private type  name arglist body
#   - creates a private proc that is only available to other procs
#     defined in the same scope. Type checking is done whenever the
#     return command is invoked within a private proc.
# private var?=value? ?,var=value....?
#   - create variables in the file scope, automatically made available
#     to any procs defined by bkproc in the same file
# private name
#   - returns the fully qualified path to name; name may be a private
#     variable or procedure.
# push list value
#   - pushes value onto a list
# pop list
#   - pops value off of list
# unless <condition> <body>
#   - equivalent to "if <condition> <do nothing> else <body>
# for <c-style expression> <body>
#   - overloads the for statement; with four args works like the normal
#     tcl for statement; with two args, first arg is treated like
#     a c-style for statement (e.g. for {j=1; $j<10; ++j} body)
#     ++foo is supported, foo++ is not

# example:
# struct S {
#    int     fee=1         // the fee to use foo
#    string  fie = "foe"   // fie is set to "foe"
#    int     foe           // foe defaults to empty string
#    list    fum={a b c}   // fum is set to list {a b c}
#    array   foo={a 1 b 2} // foo(1)=1, foo(b)=2, etc
# }
# set s [new S]
proc struct {name body {recreate "no"}} \
{
	global struct

	if {[info exists struct($name)] && $recreate ne "recreate"} {
		return -code error "illegal redefinition of struct"
	}

	set struct($name) {}
	set body [string trim $body]
	set l 0
	set r {\s*(array|float|int|index|list|string)\s+(\w+)(?:\s*=\s*(.*))?$}
	foreach line [split $body \n] {
		incr l
		set line [string trim $line]
		if {[string length $line] == 0} continue

		# a real parser would be nice, but this is good enough for now
		regsub {\s+//.*?} $line {} line ;# remove comments
		if {[regexp $r $line -- t v i]} {
			set value $i
			set type $t
			set varname $v
			# this trims pairs of leading/trailing quotes
			regsub {^\s*([\"\'])(.*)\1\s*$} $value \\2 value
			regsub {^\s*(\{(.*))\}\s*$} $value \\2 value
		} else {
			set word [scan $line %s]
			return -code error "syntax error on \
			    line $l of struct definition '$name'"
			continue
		}
		switch -exact -- $type {
			array {
				# strip surrounding {}'s
				regexp {^\{(.*)\}\s*$} $value -- value

				if {[catch {array set tmp $value} err]} {
					return -code error \
					    "value must have an even number\
					     of elements for type '$type'\
					     on line $l of struct definition\
					     '$name'"
				}
			}

		        float -
			index {
				if {![string is double $value]} {
					return -code error \
					    "'$value' is not of type '$type'\
					     on line $l of struct definition\
					     '$name'"
				}
			}
			int {
				if {![string is integer $value]} {
					return -code error \
					    "'$value' is not of type '$type'\
					     on line $l of struct definition\
					     '$name'"
				}
			}
			list {
				if {[catch {set value [lindex $value 0]} err]} {
					return -code error \
					    "'$value' is not of type '$type'\
					     on line $l of struct definition\
					     '$name'"
				}

			}
			string {
				# do nothing; everything's a valid string
			}
			default {
				return -code error "unrecognized type '$type'\
					on line $l of struct defiition\
					'$name'"
			}
		}

		lappend struct($name) $varname $type $value
	}
	return $struct($name)
}

# Create new instance of the given struct, returning the instance name
# If no name is given, or is given as "#", a name is automatically
# generated.
set structc 0
proc new {type {name \#} } \
{
	global struct structc
	# use automatically generated name
	if {$name eq "\#"} {
		set name "::.${type}_[incr structc]"
		set i 0
		# make darn sure we don't trounce on an existing object
		while {[info exists $name]} {
			append name [incr i]
		}
	} else {
		set name "::$name"
	}

	if {[info exists $name]} {
		return -code error "'$name' already exists"
	}
	foreach {varname type value} $struct($type) {
		if {$type eq "array"} {
			array set ::${name}->$varname $value
		} else {
			set ::${name}->$varname $value
		}
	}

	return $name
}

set scope {}
proc scope {name body} \
{
	global scope errorCode errorInfo

	# if the namespace for the scope doesn't exist, create it.
	# also, create some aliases to our special versions of some
	# core commands
	if {![namespace exists ::$name]} {
		namespace eval ::$name {}
		interp alias {} ::${name}::proc {} ::scope_proc
		interp alias {} ::${name}::for {} ::scope_for
		interp alias {} ::${name}::return {} ::scope_return
	}
	# in order to allow us to do 'private variable foo="bar"', we
	# need to preprocess the scope body and escape the quotes
	# so tcl doesn't strip them out when executing 'private'
	set tmp {}
	foreach line [split $body \n] {
		if {[scan $line "%s%s" cmd subcmd] == 2} {
			if {$cmd eq "private" && $subcmd eq "var"} {
				regsub -all \" $line \\\" line
			}
		}
		# this SUCKS, but it's the only way to make sure 
		# things get fully qualified when called from a public
		# proc defined in this scope. I think there's a better
		# solution but I'm not seeing it at the moment.
		regsub -all {\[private (\w+)\]} $line "::${name}::\\1" line
		lappend tmp $line
	}
	set body [join $tmp \n]
	set body [struct_pointerto $body]

	push scope $name
	set ret [catch {namespace eval ::$name $body} result]
	pop scope
	return -code $ret -errorinfo $errorInfo -errorcode $errorCode $result
}

# returns args, with the first arg fully qualified for the current scope
# e.g.: "scope foo {scope_code bar a b c}" returns "::foo::bar a b c"
proc scope_code {args} \
{
	global scope
	set s [uplevel namespace current]
	if {[llength $args] > 0} {
		return [list namespace eval $s [join $args]]
	} else {
		return $s
	}
}

# transforms code to support pointer-to
proc struct_pointerto {body} \
{
	set regex {(\$?)(\.?\w+)->(\$?\w+)(?:\(([^\)]+)\))?}
	set i 0
	while {[regexp -indices -start $i $regex $body m0 m1 m2 m3 m4]} {
		set dollar [string range $body [lindex $m1 0] [lindex $m1 1]]
		set array [string range $body [lindex $m2 0] [lindex $m2 1]]
		set index [string range $body [lindex $m3 0] [lindex $m3 1]]
		set match [string range $body [lindex $m0 0] [lindex $m0 1]]
		if {[lindex $m4 0] >= 0} {
			# foo->bar(baz)
			set subindex [string range $body \
					  [lindex $m4 0] [lindex $m4 1]]
			set index "$index\($subindex\)"
		}

		if {[string index $array 0] eq "."} {
			# name begins with ".", no second dereference
			set var "{::${array}->$index}"
		} else {
			# name requires magic second dereference. Yucky,
			# but effective.
			set var "::\[set $array\]->$index"
		}
		if {$dollar eq "\$"} {set var "\[set $var\]"}
		set body \
		    [string replace $body [lindex $m0 0] [lindex $m0 1] $var]
		set i [expr {[lindex $m0 0] + [string length $var]}]
	}
	return $body
}

# import all variables for a given scope; a call to this proc is
# automatically added to any procs created inside a scope
proc scope_import {scope} \
{
	set scopeVars {}
	set ns "::$scope"
	set locals [uplevel info locals]
	foreach var [info vars ${ns}::*] {
		set name [namespace tail $var]
		if {[lsearch -exact $locals $var] >= 0} {
			return -code error "cannot import '$name':\
				variable already exists"
		}
		append scopeVars "[list variable $var]\n"
	}

	if {[string length $scopeVars] > 0} {
		uplevel $scopeVars
	}
}

# normally, any proc created by this proc gets exported to the global
# namespace. The proc "private" will turn the export bit off.
# This proc also adds code to the proc body to automatically import
# all scope variables
proc scope_proc {name args body {export 1}} \
{
	global scope

	if {[string match {::*} $name] || ([llength $scope] == 0)} {
		# hard-coded proc names or procs defined outside
		# of a scope don't get special treatment
		::proc $name $args $body
		return
	}

	set ns "::[lindex $scope end]"
	set name "${ns}::$name"
	set pat [string map [list * \\* ? \\? \[ \\\[ \] \\\]] $name]
	if {[llength [info commands $pat]] > 0} {
		set name [namespace tail $name]
		return -code error \
		    "cannot create proc $name in scope [lindex $scope end];\
		     proc already exists"
	}
	set body "[list scope_import [lindex $scope end]]\n$body"
	::proc $name $args $body

	# this is what automagically imports the scope private
	# variables into the current context
	if {$export} {
		namespace eval $ns \
		    [list namespace export [namespace tail $name]]
		uplevel \#0 [list namespace forget $name]
		uplevel \#0 [list namespace import $name]
	}
}

# a special version of for that accepts C-like syntax in addition to
# the standard syntax, e.g. for {i=0; $i<10; i++} ...
proc scope_for {args} \
{
	if {[llength $args] == 4} {
		# this is a normal tcl for loop; call _for instead
		set command [linsert $args 0 ::for]
	} elseif {[llength $args] == 2} {
		# this is a c-style for loop; we need to parse
		set body [lindex $args 1]
		set C [Csplit [lindex $args 0]]
		if {[llength $C] != 3} {
			return -code error \
			    "for statement requires three expressions"
		}
		set start [join [c2tcl [lindex $C 0]] ";"]
		set test [join [c2tcl [lindex $C 1]] &&]
		set next [join [c2tcl [lindex $C 2]] ";"]
		set command [list for $start $test $next $body]
		
	} else {
		return -code error "wrong number of arguments"
	}
	uplevel $command
}

# a special version of the core command that attempts some type checking
# be _very_ careful modifying this proc. 
proc scope_return {args} \
{
	global scope_type
	if {[llength $args] == 0} {
		::return -code return
	} elseif {[llength $args] != 1} {
		# user is doing something more than just returning
		# a value so we'll let nature take it's course
		set cmd [lreplace [info level 0] 0 0 ::return]
		uplevel $cmd
	}
	set cmd [namespace tail [lindex [info level -1] 0]]
	set ns [uplevel namespace current]
	set fq "${ns}::$cmd"
	set ret [lindex $args 0]
	set valid 1
	if {[info exists ::scope_type($fq)]} {
		set type $::scope_type($fq)
		switch -exact -- $type {
			string {set valid 1}
			int    {set valid [string is integer -strict $ret]}
			float  {set valid [string is double -strict $ret]}
			index  {set valid [string is double $ret]}
			void   {
				if {[string length $ret] > 0} {
					set valid 0
				}
			}
			list	{
				if {[catch {lrange $ret 0 end}]} {
					set valid 0
				}
			}
		}

		unless $valid {
			::return -code error "invalid return type; \
				'$ret' is not of type '$type'"
		}
	}
	::return -code return $ret
}

# private proc name args body
# private variable name ?,name ...? (name may be foo=bar..)
# private name  (returns fully qualified version of name)
proc private {args} \
{
	global scope scope_type
	set scope [lindex $scope end]

	if {[llength $args] == 1} {
		# caller is requesting fully qualified name
		set name [lindex $args 0]
		set ns "::[lindex $scope end]"
		return "${ns}::$name"
	}

	set type [lindex $args 0]
	set args [lrange $args 1 end]
	if {$type eq "var"} {
		set result {}
		# Csplit returns a list of lists; there should be 
		# only one sublist but we have to treat it as a list 
		# of lists
		foreach s [lindex [Csplit [join $args " "]] 0] {
			if {[regexp {^(.*?)(=(.*))?$} $s -- l b r]} {
				lappend code "variable $l $r"
				lappend result "::${scope}::$l"
			} else {
				lappend code "variable $s"
				lappend result "::${scope}::$s"
			}
		}
		namespace eval $scope [join $code \n]
		return $result

	} else {
		if {[llength $args] != 3} {
			::return -code error "wrong \# args: should be\
			\"private type name args body\""
		}
		# we're creating a proc..
		set name [lindex $args 0]
		set arglist [lindex $args 1]
		set body [lindex $args 2]
		set scope_type(::${scope}::$name) $type
		uplevel [list scope_proc $name $arglist $body 0]
		return "::${scope}::$name"
	}
}

# splits C-like compound statement into a list of statements. each
# element of the list is a list of statements; each statement is a
# list of expressions, making sure not to split on "," or ";" if they
# are in a quoted string.  example: {i=1,j=2;i<10;++i} => {i=1 j=2}
# {i<10} {++i}
proc Csplit {string} \
{
	# find expression boundaries (comma), taking into consideration
	# whether or not we are inside a quoted string
	set stmts {}
	set exprs {}
	set expr {}
	set quote ""
	# N.B. '$quote eq ""' means we are not in a quoted string
	#      '$c eq $quote' means 'is this a matching quote char'
	foreach c [split $string {}] {
		# ignore whitespace if not in a quote...
		if {[string is space $c] && $quote eq ""} continue
		if {$c eq "," && $quote eq ""} {
			# end of this sub-expression
			lappend exprs $expr
			set expr ""
		} elseif {$c eq ";" && $quote eq ""} {
			# end of subexpression, and end of expression
			if {$expr ne ""} {lappend exprs $expr}
			if {[llength $exprs] > 0} {lappend stmts $exprs}
			set expr ""
			set exprs {}
		} elseif {$c eq $quote} {
			append expr $c
			set quote ""
		} elseif {$quote eq "" && ($c eq "\"" || $c eq "\{")} {
			append expr $c
			if {$c eq "\{"} {
				set quote "\}"
			} else {
				set quote "\""
			}
		} else {
			append expr $c
		}

	}
	if {$expr ne ""} {lappend exprs $expr}
	if {[llength $exprs] > 0} {lappend stmts $exprs}

	return $stmts
}

# convert C-like expressions into tcl expressions; expects as input
# a list of C expressions. This doesn't do a very good job; instead
# of regexps I think I just need to step over the string byte by
# byte.
# Danger Will Robinson!
# this takes a list of expressions rather than a single expression; 
# typically the input to this is one element from the output of Csplit.
# Thus, 'c2tcl "a={foo bar}"' is wrong; 'c2tcl [list "a={foo bar}"]'
# is right.
proc c2tcl {list} {
	set result {}
	foreach e $list {
		if {[regexp {^\s*(.*?)\s*([-+=<>]{1,2})\s*(.*)\s*$} $e -- l op r]} {
			switch -exact -- $op {
				= {
					set q [string index $r 0]
					if {$q eq "\"" || $q eq "\{"} {
						set r [string range $r 1 end-1]
					}
					lappend result "[list set $l $r]"
				}
				eq - ne - == - != - <= - >= - < - > {
					lappend result "$l $op $r"
				}
				+= {lappend result "incr $l $r"}
				-= {lappend result "incr $l -$r"}
				++ - -- {
					if {$l eq ""} {set l $r}
					if {$op eq "--"} {
						lappend result "incr $l -1"
					} else {
						lappend result "incr $l 1"
					}
				}
				default {
					lappend result "$l $op $r"
				}
			}
		} else {
			lappend result $e
		}
	}
	return $result
}

proc unless {expr command} \
{
	global errorInfo errorCode

	set code [catch {uplevel [list if $expr {} else $command]} result]
	switch -exact -- $code {
		0	{return $result}
		1	{return -errorinfo $errorInfo -errorcode $errorCode \
			     -code error $result}
		2	{return -code return $result}
		3	{return -code break}
		4	{return -code continue}
		default {return -code $code $result}
	}
}

proc push {varname args} \
{
	upvar $varname list
	foreach value $args {
		lappend list $value
	}
	return $list
}

proc pop {varname} \
{
	upvar $varname list
	set last [lindex $list end]
	set list [lrange $list 0 end-1]
	return $last
}

